/* -*- mode: c++; tab-width: 8; c-file-style: "gnu" -*- */
/*
  SDL - Simple DirectMedia Layer
  Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002  Sam Lantinga

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Sam Lantinga
  slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysvideo.cc,v 1.3 2002/09/01 00:37:24 slouken Exp $";
#endif

/* Qtopia based framebuffer implementation */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
extern int errno;

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fb.h>
#include <sys/mman.h>
#include <asm/page.h>	

#include <qapplication.h>
#include <qpe/qpeapplication.h>
#include <qpe/qcopenvelope_qws.h>
#include <qgfx_qws.h>
#include <qwindowsystem_qws.h>
#include <qwidget.h>
#include <qwidgetlist.h>
#include <qdirectpainter_qws.h>

#include "SDL.h"
#include "SDL_timer.h"

#include "SDL_bvddQWin.h"
#include "SDL_sysvideo.h"

extern "C" {

#include "SDL_sysvideo.h"
#include "SDL_bvddsysmouse_c.h"
#include "SDL_bvddsysevents_c.h"
#include "SDL_events_c.h"
#include "SDL_bvddsyswm_c.h"
#include "SDL_bvddlowvideo.h"
#include "bvdd.h"
#define uint32_t unsigned int

  // #define QTOPIA_DEBUG
#define QTOPIA_LOG
#define QT_HIDDEN_SIZE	32	/* starting hidden window size */

#ifdef QTOPIA_LOG
  static FILE *st_logfp = NULL;
#endif
  static inline void LOG(char *fmt, ...)
  {
#ifdef QTOPIA_LOG
    va_list ap;

    va_start(ap, fmt);
    vfprintf(st_logfp, fmt, ap);
    if (st_logfp != stderr) {
      fflush(st_logfp);
      vfprintf(stderr, fmt, ap);
    }
    va_end(ap);
#endif
  }


/* Name of the environment variable used to invert the screen rotation or not:
   Possible values:
   !=0 : Screen is 270- rotated
   0: Screen is 90- rotated*/
#define SDL_QT_ROTATION_ENV_NAME "SDL_QT_ROTATION"
#define SDL_QT_FORCE_QVGA "SDL_FORCE_QVGA"

  /* Initialization/Query functions */
  static int QT_VideoInit(_THIS, SDL_PixelFormat *vformat);
  static SDL_Rect **QT_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
  static SDL_Surface *QT_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
  static void QT_UpdateMouse(_THIS);
  static int QT_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
  static void QT_VideoQuit(_THIS);

  /* Hardware surface functions */
  static int QT_AllocHWSurface(_THIS, SDL_Surface *surface);
  static int QT_LockHWSurface(_THIS, SDL_Surface *surface);
  static void QT_UnlockHWSurface(_THIS, SDL_Surface *surface);
  static void QT_FreeHWSurface(_THIS, SDL_Surface *surface);

  static int QT_ToggleFullScreen(_THIS, int fullscreen);

  static int QT_IconifyWindow(_THIS);
  static SDL_GrabMode QT_GrabInput(_THIS, SDL_GrabMode mode);

  static int bvdd_fd;

#define LCD_QVGA 0
#define LCD_VGA  1

  static short *gfx_buffer;
  static int gfx_width;
  static int gfx_height;

  static short *bvdd_buffer;
  static int bvdd_width;
  static int bvdd_height;


  static int lcd_mode;

  static int isKHloaded;
 
  /* FB driver bootstrap functions */

  static int QT_Available(void)
  {
    return(1);
  }

  static void QT_DeleteDevice(SDL_VideoDevice *device)
  {
    free(device->hidden);
    free(device);
  }

  static SDL_VideoDevice *QT_CreateDevice(int devindex)
  {
    SDL_VideoDevice *device;

    /* Initialize all variables that we clean on shutdown */
    device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
    if ( device ) {
      memset(device, 0, (sizeof *device));
      device->hidden = (struct SDL_PrivateVideoData *)
	malloc((sizeof *device->hidden));
    }
    if ( (device == NULL) || (device->hidden == NULL) ) {
      SDL_OutOfMemory();
      if ( device ) {
	free(device);
      }
      return(0);
    }
    memset(device->hidden, 0, (sizeof *device->hidden));

    /* Set the function pointers */
    device->VideoInit = QT_VideoInit;
    device->ListModes = QT_ListModes;
    device->SetVideoMode = QT_SetVideoMode;
    device->UpdateMouse = QT_UpdateMouse;
    device->SetColors = QT_SetColors;
    device->UpdateRects = NULL;
    device->VideoQuit = QT_VideoQuit;
    device->AllocHWSurface = QT_AllocHWSurface;
    device->CheckHWBlit = NULL;
    device->FillHWRect = NULL;
    device->SetHWColorKey = NULL;
    device->SetHWAlpha = NULL;
    device->LockHWSurface = QT_LockHWSurface;
    device->UnlockHWSurface = QT_UnlockHWSurface;
    device->FlipHWSurface = NULL;
    device->FreeHWSurface = QT_FreeHWSurface;
    device->SetIcon = NULL;
    device->SetCaption = QT_SetWMCaption;
    device->IconifyWindow = QT_IconifyWindow;
    device->GrabInput = QT_GrabInput;
    device->GetWMInfo = NULL;
    device->FreeWMCursor = QT_FreeWMCursor;
    device->CreateWMCursor = QT_CreateWMCursor;
    device->ShowWMCursor = QT_ShowWMCursor;
    device->WarpWMCursor = QT_WarpWMCursor;
    device->InitOSKeymap = QT_InitOSKeymap;
    device->PumpEvents = QT_PumpEvents;

    device->free = QT_DeleteDevice;
    device->ToggleFullScreen = QT_ToggleFullScreen;

    /* Set the driver flags */
    device->handles_any_size = 0;
	
    return device;
  }

  VideoBootStrap Qtopia_bootstrap = {
    "bvdd", "BVDD overlay graphics",
    QT_Available, QT_CreateDevice
  };

  /* Function to sort the display_list */
  static int CompareModes(const void *A, const void *B)
  {
    return 0;
  }

  /* Yes, this isn't the fastest it could be, but it works nicely */
  static int QT_AddMode(_THIS, int index, unsigned int w, unsigned int h)
  {
    SDL_Rect *mode;
    int i;
    int next_mode;

    /* Check to see if we already have this mode */
    if ( SDL_nummodes[index] > 0 ) {
      for ( i=SDL_nummodes[index]-1; i >= 0; --i ) {
	mode = SDL_modelist[index][i];
	if ( (mode->w == w) && (mode->h == h) ) {
	  return(0);
	}
      }
    }

    /* Set up the new video mode rectangle */
    mode = (SDL_Rect *)malloc(sizeof *mode);
    if ( mode == NULL ) {
      SDL_OutOfMemory();
      return(-1);
    }
    mode->x = 0;
    mode->y = 0;
    mode->w = w;
    mode->h = h;
#ifdef QTOPIA_DEBUG
    fprintf(stderr, "Adding mode %dx%d at %d bytes per pixel\n", w, h, index+1);
#endif

    /* Allocate the new list of modes, and fill in the new mode */
    next_mode = SDL_nummodes[index];
    SDL_modelist[index] = (SDL_Rect **)
      realloc(SDL_modelist[index], (1+next_mode+1)*sizeof(SDL_Rect *));
    if ( SDL_modelist[index] == NULL ) {
      SDL_OutOfMemory();
      SDL_nummodes[index] = 0;
      free(mode);
      return(-1);
    }
    SDL_modelist[index][next_mode] = mode;
    SDL_modelist[index][next_mode+1] = NULL;
    SDL_nummodes[index]++;

    return(0);
  }

  int QT_VideoInit(_THIS, SDL_PixelFormat *vformat)
  {
    const char *SDL_bvdddev;

     /* Initialize the QPE Application  */
     /* Determine the screen depth */
    vformat->BitsPerPixel = QPixmap::defaultDepth();

    // For now we hardcode the current depth because anything else
    // might as well be emulated by SDL rather than by Qtopia.

    //frame buffer device open.
    SDL_bvdddev = getenv("SDL_BVDDDEV");
    if ( SDL_bvdddev == NULL ) {
    	SDL_bvdddev = "/dev/bvdd";
    }
    bvdd_fd = open(SDL_bvdddev, O_RDWR, 0);
    if ( bvdd_fd < 0 ) {
    	SDL_SetError("Unable to open %s", SDL_bvdddev);
    	return(-1);
    }
    close(bvdd_fd);


    QSize desktop_size = qApp->desktop()->size();
    QT_AddMode(_this, ((vformat->BitsPerPixel+7)/8)-1,
	       desktop_size.width(), desktop_size.height());
    QT_AddMode(_this, ((vformat->BitsPerPixel+7)/8)-1,
	       desktop_size.height(), desktop_size.width());

    /* Create the window / widget */
    SDL_Win = new SDL_QWin(QSize(QT_HIDDEN_SIZE, QT_HIDDEN_SIZE));
    ((QPEApplication*)qApp)->showMainWidget(SDL_Win);
    /* Fill in some window manager capabilities */
    _this->info.wm_available = 0;

#ifdef QTOPIA_LOG
    st_logfp = fopen("/tmp/sdl-qt-debug", "w");
    if (!st_logfp)
      st_logfp = stderr;
#endif

    QT_GrabInput(_this, SDL_GRAB_ON);

    // keyhelper loading for SDL

    isKHloaded=(fopen("/home/zaurus/Settings/keyhelper_SDL.xml","r") != NULL) ? 1 : 0;
    if (isKHloaded) {
	system("qcop QPE/KeyHelper \"reload(QString)\" keyhelper_SDL.xml"); //QcopEnvelope can't use. 
	LOG("keyhelper_SDL.xml loaded\n");
    }
    return(0);
  }

  /* We support any dimension at our bit-depth */
  SDL_Rect **QT_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
  {
    SDL_Rect **modes;

    modes = ((SDL_Rect **)0);
    if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
      modes = SDL_modelist[((format->BitsPerPixel+7)/8)-1];
    } else {
      if ( format->BitsPerPixel ==
	   _this->screen->format->BitsPerPixel ) {
	modes = ((SDL_Rect **)-1);
      }
    }
    return(modes);
  }

  /* Various screen update functions available */
  static void QT_NormalUpdate0(_THIS, int numrects, SDL_Rect *rects);
  static void QT_NormalUpdate3(_THIS, int numrects, SDL_Rect *rects);
  static void QT_DirectUpdate(_THIS, int numrects, SDL_Rect *rects);


  static int QT_SetFullScreen(_THIS, SDL_Surface *screen, int fullscreen)
  {
    return -1;
  }

  static int QT_ToggleFullScreen(_THIS, int fullscreen)
  {
    return -1;
  }

  static void QT_GetQteServerSpec(_THIS, int *rotation, bool *isQvga)
  {
    const char *user;
    char buf[FILENAME_MAX];
    FILE *fp;
    int rot = 0, is_qvga = 0;

    user = getenv("USER");
    snprintf(buf, sizeof(buf), "/tmp/qtembedded-%s/QtEmbedded-0.spec",
             user ? user : "root");
    LOG("QT_GetRotation: Read spec from '%s'\n", buf);

    fp = fopen(buf, "rb");
    if (fp) {
      int size;
      if ((size = fread(buf, 1, sizeof(buf) - 1, fp)) > 0) {
        buf[size] = '\0';
        LOG(" spec is '%s'\n", buf);
        // get rotation value
        if (strstr(buf, "Rot")) {
          rot = atoi(strstr(buf, "Rot") + 3);
          rot /= 90;
          if (rot < 0 || rot > 3) {
            rot = 0;
          }
        } else
          rot = 0;

        // get qvga mode in SL-C700
        if (strstr(buf, "Qvga"))
          is_qvga = 1;
      }
      fclose(fp);
    }

    LOG(" Rot=%d, Qvga=%d\n", rot, is_qvga);

    if (rotation)
      *rotation = rot;
    if (isQvga)
      *isQvga = is_qvga;
  }



  /* FIXME: check return values and cleanup here */
  SDL_Surface *QT_SetVideoMode(_THIS, SDL_Surface *current,
			       int width, int height, int bpp, Uint32 flags)
  {
    QImage *qimage;
    QSize qteSize = qApp->desktop()->size();
    QSize fbSize;
    QSize userSize;
    int tmp_ioctl_data;

    void *mapped_mem;
    int vram_size = 0;

    int qteRotation, userRotation, sdlRotation=0;
    int qteKeyRotation, sdlKeyRotation;
    bool isQteQvga;
    bool force_qvga;

    // qte Ǥβž٤
    QT_GetQteServerSpec(_this, &qteRotation, &isQteQvga);

    // SL-C700 Υ
    bool isInputStyle = false;
    int status = system("/home/QtPalmtop/bin/chkhinge");
    if (WEXITSTATUS(status) != 2)
      isInputStyle = true;
    LOG("QT_SetVideoMode: SL-C700 Style is %s\n",
	isInputStyle ? "Input style" : "View style");

    if(isInputStyle)
      qteRotation = 3;

    //
    const char *envQVGA = getenv(SDL_QT_FORCE_QVGA);
    if(envQVGA){
      force_qvga = atoi(envQVGA) == 1 ? TRUE : FALSE;
    }else{
      force_qvga = FALSE;
    }

    const char *envRotStr = getenv(SDL_QT_ROTATION_ENV_NAME);
    if(envRotStr){
      userRotation = atoi(envRotStr);
      sdlRotation = userRotation >= -1 ? userRotation : qteRotation;
    }else {
      userRotation = 0;
      sdlRotation = qteRotation;
    }

    LOG("QT_SetVideoMode: argSize=%dx%d\n", width, height);
    LOG("QT_SetVideoMode: qteSize=%dx%d\n", qteSize.width(), qteSize.height());
    LOG("QT_SetVideoMode: qteRotation=%d\n", qteRotation);
    LOG("QT_SetVideoMode: userRotation=%d\n", userRotation);
    LOG("QT_SetVideoMode: sdlRotation=%d\n", sdlRotation);
    
    
    
    // keyboard rotation
    qteKeyRotation = qteRotation;
    if (isQteQvga)
      qteKeyRotation = 3;
    
    sdlKeyRotation = qteKeyRotation;
    
    SDL_Win->setKeyRotation(sdlKeyRotation, qteKeyRotation);
    
    LOG("QT_SetVideoMode: qteKeyRotation=%d\n", qteKeyRotation);
    LOG("QT_SetVideoMode: sdlKeyRotation=%d\n", sdlKeyRotation);
    
    if ( flags & SDL_OPENGL ) {
      SDL_SetError("OpenGL not supported");
      return(NULL);
    }
    /* Create the QImage framebuffer */

    bvdd_fd = open("/dev/bvdd", O_RDWR);
    if (bvdd_fd < 0) {
	perror("open");
	exit(1) ;
    }


    current->flags = 0;//SDL_FULLSCREEN; // We always run fullscreen.

    if(userRotation == -1){
      current->w = fbSize.width();
      current->h = fbSize.height();
    }else{
      current->w = width;
      current->h = height;
    }
    
    if( ((width <=240) && ( height <=320) && ( sdlRotation == 0 || sdlRotation == -1) ) || 
	((width <=320) && ( height <=240) && ( sdlRotation == 3 ) ) || 
	force_qvga){
      fbSize.setWidth(240);
      fbSize.setHeight(320);
      ioctl(bvdd_fd, BVDD_IOCTL_CHANGE_QVGA);
      lcd_mode = LCD_QVGA;
      vram_size = 240 * 320 * 2;
    }else{
      fbSize.setWidth(480);
      fbSize.setHeight(640);
      lcd_mode = LCD_VGA;
      vram_size = 480 * 640 * 2;
    }

    if (ioctl(bvdd_fd, BVDD_IOCTL_SETUP_VRAM, vram_size) < 0) {
	perror("ioctl(BVDD_IOCTL_SETUP_VRAM)");
	close(bvdd_fd);
	exit(1);
    }


    SDL_Win->setScreenRotation(sdlRotation, qteRotation);
    SDL_Win->setFBSize(fbSize);

    {
	bvdd_frame_t frame;

	frame.format = BVDD_FORMAT_RGB16;
	if(lcd_mode == LCD_QVGA){
	  frame.width = 240;
	  frame.height = 320;
	}else{
	  frame.width = 480;
	  frame.height = 640;
	}
	bvdd_width = frame.width;
	bvdd_height = frame.height;

	frame.x = 0;
	frame.y = 0;
	frame.offset = BVDD_VRAM_OFFSET_USER;
	{
	  FILE *fp;

	  fp = fopen("/proc/zaurus/CCCR", "r");
	  if (fp) {
	    unsigned int cccr, cccr_orig;
	    unsigned int vcore, vcore_orig;
	    char command[100];
	    
	    fclose(fp);
	    
	    cccr = 0x02000210;
	    vcore = 0x1a;
	    ioctl(bvdd_fd, BVDD_IOCTL_GET_CCCR, &cccr_orig);
	    ioctl(bvdd_fd, BVDD_IOCTL_GET_VCORE, &vcore_orig);
	    
	    if (cccr_orig != 0 && cccr_orig != 0x2000210){
	      if (vcore_orig < 0x1a)
		ioctl(bvdd_fd, BVDD_IOCTL_SET_VCORE, vcore);
	      
	      ioctl(bvdd_fd, BVDD_IOCTL_SET_CCCR, cccr);
	    }
	    
	    ioctl(bvdd_fd, BVDD_IOCTL_SETUP_BASE_FRAME, &frame);
	    
	    if (cccr_orig != 0 && cccr_orig != 0x2000210){
	      ioctl(bvdd_fd, BVDD_IOCTL_SET_CCCR, cccr_orig);
	      ioctl(bvdd_fd, BVDD_IOCTL_SET_VCORE, vcore_orig);
	    }
	    
	    
	    LOG("frame.format = %d\n",frame.format);
	    LOG("frame.width = %d\n",frame.width);
	    LOG("frame.height = %d\n",frame.height);
	    
	  }else {
	    ioctl(bvdd_fd, BVDD_IOCTL_SETUP_OVERLAY1_FRAME, &frame);
	  }
	}
    }
    

    mapped_mem = mmap(0, vram_size, PROT_READ|PROT_WRITE, MAP_SHARED, bvdd_fd, BVDD_VRAM_OFFSET_USER);
    if (mapped_mem == MAP_FAILED) {
      LOG("BVDD map failed size %d. %x error no= %d\n",vram_size, mapped_mem, errno);
      exit(1);
    }
    
    LOG("BVDD maped size %d address %x.\n", vram_size, mapped_mem);
    
    short *fill = (short*)mapped_mem;
    for (int i = 0; i <(vram_size/2); i++)
      *fill++ = 0;
	    


    if(sdlRotation == -1){
      current->pitch = current->w * 2;
    }else{
      current->pitch = width * 2;
      gfx_buffer = (short *)malloc(width * height * 2);
      gfx_width = width;
      gfx_height = height;
    }

    current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE);
    
    if(sdlRotation == -1 ) {
      current->pixels = (void *)(mapped_mem);
      _this->UpdateRects = QT_DirectUpdate;
      LOG("Using DirectPaint mode\n");
    } else if(sdlRotation == 0){
      current->pixels = (void *)gfx_buffer;
      bvdd_buffer = (short *)(mapped_mem);
      _this->UpdateRects = QT_NormalUpdate0;
      LOG("Using Rotation 0\n");
    } else { // if(sdlRotation == 3){
      current->pixels = (void *)gfx_buffer;
      bvdd_buffer = (short *)(mapped_mem);
      _this->UpdateRects = QT_NormalUpdate3;
      LOG("Using Rotation 270\n");
    }
    

    //     fprintf(stderr,"QT_SetVideoMode() qImage:%dx%d %d\n",
    // 			qimage->width(), qimage->height(), qimage->bytesPerLine());
    /* We're done */
    return(current);
  }


  /* Update the current mouse state and position */
  void QT_UpdateMouse(_THIS)
  {
    QPoint point(-1, -1);
    if ( SDL_Win->isActiveWindow() ) {
      point = SDL_Win->mousePos();
    }
    
    if ( (point.x() >= 0) && (point.x() < SDL_VideoSurface->w) &&
	 (point.y() >= 0) && (point.y() < SDL_VideoSurface->h) ) {
      SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
      SDL_PrivateMouseMotion(0, 0,
			     (Sint16)point.x(), (Sint16)point.y());
    } else {
      SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);
    }
  }

  /* We don't actually allow hardware surfaces other than the main one */
  static int QT_AllocHWSurface(_THIS, SDL_Surface *surface)
  {
    return(-1);
  }
  static void QT_FreeHWSurface(_THIS, SDL_Surface *surface)
  {
    return;
  }
  static int QT_LockHWSurface(_THIS, SDL_Surface *surface)
  {
    return(0);
  }
  static void QT_UnlockHWSurface(_THIS, SDL_Surface *surface)
  {
    return;
  }

  static void QT_NormalUpdate0(_THIS, int numrects, SDL_Rect *rects)
  {
    for(int i=0; i<numrects; ++i ) {
      short *srcBits, *dstBits;
      int w,h;
      srcBits = gfx_buffer + rects[i].x + rects[i].y * gfx_width;
      dstBits = ( bvdd_buffer 
		  + ((bvdd_width - gfx_width) >> 1) 
		  + ((bvdd_height - gfx_height) >> 1) * bvdd_width
		  + rects[i].x + rects[i].y * bvdd_width );
      w = rects[i].w << 1;
      h = rects[i].h;

      while (h--) {
	memcpy(dstBits, srcBits, w);
	dstBits += bvdd_width;
	srcBits += gfx_width;
      }
    }
  }

  static void QT_NormalUpdate3(_THIS, int numrects, SDL_Rect *rects)
  {
    for(int i=0; i<numrects; ++i ) {

      if (rects[i].w != bvdd_width || rects[i].h != bvdd_height) {
	// partial update
	const short *pSrc = gfx_buffer + (rects[i].x + rects[i].y * gfx_width);
	int dstBorderLeftRight = (bvdd_width - gfx_height) >>1;
	int dstBorderTopBottom = (bvdd_height - gfx_width) >>1;
	int dstStartOffsetX = bvdd_width - dstBorderLeftRight - 1;
	int dstStartOffsetY = dstBorderTopBottom;
	short *pDst = bvdd_buffer +
	  (dstStartOffsetY * bvdd_width + dstStartOffsetX);
	pDst += rects[i].x * bvdd_width;
	pDst -= rects[i].y;
	
	int width = rects[i].w, height = rects[i].h, w;
	int srcYAdd = gfx_width - rects[i].w;
	int dstXAdd = bvdd_width;
	int dstYSub = (bvdd_width * rects[i].w) + 1;
	
	
	while (height--) {
	  w = width;
	  while (w--) {
	    *pDst = *pSrc ++;
	    pDst += dstXAdd;
	  }
	  pSrc += srcYAdd;
	  pDst -= dstYSub;
	}
	// 		fprintf(stderr, "---- Blit end\n");
      } else {
	// full update
	const short *src = gfx_buffer;
	short *dst = bvdd_buffer + (bvdd_width - 1);
	int w = rects[i].w;
	int h = rects[i].h;
	int i, dstSub = bvdd_width * w + 1;
	int dstAdd = bvdd_width;
	
	while (h--) {
	  i = w;
	  while (i--) {
	    *dst = *src ++;
	    dst += dstAdd;
	  }
	  dst -= dstSub;
	}
      }

    }

  }


  static void QT_DirectUpdate(_THIS, int numrects, SDL_Rect *rects)
  {

  }

  /* Is the system palette settable? */
  int QT_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
  {
    return -1;
  }

  void QT_VideoQuit(_THIS)
  {
    // This is dumb, but if I free this, the app doesn't exit correctly.
    // Of course, this will leak memory if init video is done more than once.
    // Sucks but such is life.
    
    //    -- David Hedbor
    //    delete SDL_Win; 
    //    SDL_Win = 0;
    printf("exit SDL\n");
    if(lcd_mode == LCD_QVGA)
      ioctl(bvdd_fd, BVDD_IOCTL_CHANGE_VGA);

    if ( bvdd_fd > 0 ) {
    	/* Restore the original video mode and palette */
    	/* We're all done with the framebuffer */
    	close(bvdd_fd);
    	bvdd_fd = -1;
    }

    _this->screen->pixels = NULL;
    QT_GrabInput(_this, SDL_GRAB_OFF);

    SDL_Win->toggleSuspend(true);

/*
    {
      QCopEnvelope e("QPE/KeyHelper", "repeater(int)");
      e << 1;
    }
*/
    if (isKHloaded)
	system("qcop QPE/KeyHelper \"reload()\""); //QcopEnvelope can't use. 

  }



  static int QT_IconifyWindow(_THIS) {
    SDL_Win->hide();
    
    return true;
  }

  static SDL_GrabMode QT_GrabInput(_THIS, SDL_GrabMode mode) {
    if(mode == SDL_GRAB_OFF) {
      QPEApplication::grabKeyboard();
      qApp->processEvents();
      QPEApplication::ungrabKeyboard();
    } else {
      QPEApplication::grabKeyboard();
    }
    qApp->processEvents();
    return mode;
  }

}; /* Extern C */
